/*
 * Copyright 2023, Exotic Silicon, all rights reserved.
 *
 * For more information about this software, please visit:
 * https://research.exoticsilicon.com/articles/utf8_parsing_in_c
 * gemini://gemini.exoticsilicon.com/articles/utf8_parsing_in_c
 *
 * This is example benchmarking code to demonstrate the relative performance of different routines that each perform the same count
 * of leading set bits across a buffer of random or pseudo-random data.  Such counting of leading set bits has applications in
 * handling UTF-8 encoded text streams.
 *
 * Seven versions of the function are presented, four written in C and three written in X86-64 assembler.
 *
 * Compilation with a selection of compiler optimisation flags is possible and will give different results.  Typically -O3 is a good
 * starting point.
 *
 * The default buffer size is 128 Mb, but this can be changed by adjusting BUFFER_SIZE.
 *
 * By default, the buffer is filled with random byte values 0x00 - 0xFF.
 * To bias the random data towards ASCII values 0x20 - 0x7F, move the line below which defines SIMULATE_MOSTLY_ASCII outside of this
 * comment block.
 *
 * #define SIMULATE_MOSTLY_ASCII
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. This software is licensed exclusively under this specific license text.  The license text may not be changed, and the software
 *    including modified versions may not be re-licensed under any other license text.
 * 2. Redistributions of source code must retain the above copyright notice, this list of conditions, and the following disclaimer.
 * 3. Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the following
 *    disclaimer in the documentation and/or other materials provided with the distribution.
 * 4. All advertising materials mentioning features or use of this software must display the following acknowledgement: This product
 *    includes software developed by Exotic Silicon.
 * 5. The name of Exotic Silicon must not be used to endorse or promote products derived from this software without specific prior
 *    written permission.
 * 6. Redistributions of modified versions of the source code must be clearly identified as having been modified from the original.
 * 7. Redistributions in binary form that have been created from modified versions of the source code must clearly state in the
 *    documentation and/or other materials provided with the distribution that the source code has been modified from the original.
 *
 * THIS SOFTWARE IS PROVIDED 'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EXOTIC SILICON BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Version 1.0
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <string.h>

#define BUFFER_SIZE 1 << 27
#define NUMBER_OF_ALGOS 7

int main()
{
unsigned char * inbuf;
unsigned int i;
unsigned int j;
unsigned int bits[9];
unsigned int known_good[9];
unsigned int leading_ones;
unsigned int bitops;
unsigned int masks[9];
unsigned int method;
unsigned int flag_first_run;
unsigned char * method_names[]={"C explicit checks", "C bitshifting", "C bit checking", "C bit pre-masking", "ASM bitshifting only", "ASM entire algo", "ASM 64-bit reads"};
unsigned int pass_number;
struct timespec best_t[NUMBER_OF_ALGOS];
struct timespec start_t;
struct timespec end_t;
struct timespec elapsed_t;

for (i=0; i < NUMBER_OF_ALGOS; i++) {
	best_t[i].tv_sec = 1000000;
	best_t[i].tv_nsec = 0;
	}

inbuf=malloc((BUFFER_SIZE)+8);
if (inbuf==NULL) {
	printf ("Memory allocation failed\n");
	return (1);
	}

#ifndef SIMULATE_MOSTLY_ASCII
arc4random_buf(inbuf, BUFFER_SIZE);
#endif

printf ("Size of source buffer containing random data is %u bytes.\n", BUFFER_SIZE);

/*
 * Pre-compute the bitmasks for the pre-computed bitchecking algorithm.
 */
for (i=0; i<8; i++) {
	masks[i]=(0x80 >> i);
	}
masks[8]=0;

#ifdef SIMULATE_MOSTLY_ASCII
for (i=0; i < BUFFER_SIZE; i++) {
	if (arc4random_uniform(32)>1) {
		*(inbuf+i)=arc4random_uniform(96)+32;
		}
	}
#endif

flag_first_run=0;
for (pass_number=0; pass_number < 50; pass_number++) {
	method=arc4random_uniform(NUMBER_OF_ALGOS);
	for (i=0; i < 9; i++) {
		bits[i]=0;
		}

	printf ("Starting %s... \n", method_names[method]);

	if (method==1) {
		clock_gettime(CLOCK_REALTIME, &start_t);
		for (i=0; i < BUFFER_SIZE; i++) {
			bitops=inbuf[i];
			for (leading_ones=0; (bitops & 0x80)==0x80; leading_ones++, bitops=bitops << 1) ;
			bits[leading_ones]++;
			}
		clock_gettime(CLOCK_REALTIME, &end_t);
		}
	if (method==2) {
		clock_gettime(CLOCK_REALTIME, &start_t);
		for (i=0; i < BUFFER_SIZE; i++) {
			bitops=inbuf[i];
			for (leading_ones=0; (bitops & (0x80 >> leading_ones)) != 0; leading_ones++) ;
			bits[leading_ones]++;
			}
		clock_gettime(CLOCK_REALTIME, &end_t);
		}
	if (method==3) {
		clock_gettime(CLOCK_REALTIME, &start_t);
		for (i=0; i < BUFFER_SIZE; i++) {
			bitops=inbuf[i];
			for (leading_ones=0; (bitops & masks[leading_ones]) != 0; leading_ones++) ;
			bits[leading_ones]++;
			}
		clock_gettime(CLOCK_REALTIME, &end_t);
		}
	if (method==4) {
		clock_gettime(CLOCK_REALTIME, &start_t);
		/*
		 * Optimised assembler code using lzcnt for the actual bit counting _only_.
		 * Looping over the input and writing the results to the bits[] array is still done in C.
		 */
		for (i=0; i < BUFFER_SIZE; i++) {
			bitops=inbuf[i];
			asm volatile ("shl $24, %0; not %0; lzcntl %0, %1" : "+r" (bitops) ,"+r" (leading_ones)::);
			bits[leading_ones]++;
			}
		clock_gettime(CLOCK_REALTIME, &end_t);
		}
	if (method==5) {
		clock_gettime(CLOCK_REALTIME, &start_t);
		/*
		 * Optimised assembler code to do the whole process.
		 * This includes the loop over the input and writing the results to the bits[] array.
		 * However this version still does it one byte at a time, doing a memory fetch for each byte of the input.
		 */
		asm volatile (
		    "loop:; dec %2; mov (%0,%2,1), %%rcx; shl $56, %%rcx; not %%rcx; lzcntq %%rcx, %%r15;"
		    "leaq(%1,%%r15,4), %%r14;"
		    "incl (%%r14); test %2, %2; jnz loop;"
		     :: "r" (inbuf), "r" (bits), "r" ((uint64_t)BUFFER_SIZE) : "rcx", "r15", "r14", "memory");
		clock_gettime(CLOCK_REALTIME, &end_t);
		}
	if (method==6) {
		clock_gettime(CLOCK_REALTIME, &start_t);
		/*
		 * Do 64-bit reads from memory and operate on 8-byte blocks.
		 */
		asm volatile (
		    "mov %1, %%r13; add $32, %%r13;"		/* Store the address of bits[8] in %r13 for quick access later */
		    "shrq $3, %2;"
		    "outerloop:"
		    "dec %2; mov (%0,%2,8), %%rdx;"		/* %rdx now contains 64 bits of useful data */
		    "not %%rdx;"				/* Only invert once for the whole 64 bits */
		    "mov $8, %%rcx;"				/* %rcx is the inner loop counter */
		    "innerloop:"
		    "lzcntq %%rdx, %%r15;"			/* Count leading zeros, might go beyond 8 and into the next byte */
		    "cmp $8, %%r15;"				/* Test for 8 */
		    "jl skip;"					/* If < 8 then skip forward to the regular handling code */
		    "incl (%%r13);"				/* Otherwise, hard-code an increase to bits[8]... */
		    "shl $8, %%rdx;"				/* ... and unroll the loop code for speed */
		    "dec %%rcx; test %%rcx, %%rcx;"
		    "jnz innerloop;"
		    "test %2, %2; jnz outerloop;jmp out;"
		    "skip: leaq(%1,%%r15,4), %%r14;"
		    "incl (%%r14);"				/* Forward one byte in the output */
		    "shl $8, %%rdx;"				/* Move to the next byte of the 8 in the inner loop */
		    "dec %%rcx; test %%rcx, %%rcx;"		/* Check inner loop counter */
		    "jnz innerloop;"				/* Repeat inner loop */
		    "test %2, %2; jnz outerloop;out:"
		     :: "r" (inbuf), "r" (bits), "r" ((uint64_t)BUFFER_SIZE) : "rcx", "rdx", "r15", "r14", "r13", "memory");
		clock_gettime(CLOCK_REALTIME, &end_t);
		}
	if (method==0) {
		clock_gettime(CLOCK_REALTIME, &start_t);
		for (i=0; i < BUFFER_SIZE; i++) {
			bitops=inbuf[i];
			if (bitops < 0x80) {
				leading_ones=0;
				} else {
				if (bitops < 0xC0) {
					leading_ones=1;
					} else {
					if (bitops < 0xE0) {
						leading_ones=2;
						} else {
						if (bitops < 0xF0) {
							leading_ones=3;
							} else {
							if (bitops < 0xF8) {
								leading_ones=4;
								} else {
								if (bitops < 0xFC) {
									leading_ones=5;
									} else {
									if (bitops < 0xFE) {
										leading_ones=6;
										} else {
										if (bitops == 0xFE) {
											leading_ones=7;
											} else {
											leading_ones=8;
											}
										}
									}
								}
							}
						}
					}
				}
			bits[leading_ones]++;
			}
		clock_gettime(CLOCK_REALTIME, &end_t);
		}

	timespecsub(&end_t, &start_t, &elapsed_t);

	if (flag_first_run != 0) {
		for (i=0; i<9; i++) {
#ifdef DEBUG
			printf ("Method %d produced output %d/%d\n", method, known_good[i], bits[i]);
#endif
			if (known_good[i] != bits[i]) {
				printf ("Fatal error - method %d produced non-consistent output %d/%d!\n", method, known_good[i], bits[i]);
				printf ("Check compiler options and hardware.\n");
				return (1);
				}
			}
#ifdef DEBUG
		printf ("Results match original method.\n");
#endif
		}

	if (flag_first_run==0) {
		for (i=0; i<9; i++) {
			known_good[i]=bits[i];
			}
		flag_first_run=1;
		printf ("Results stored for comparison.\n");
		}

	printf ("Runtime: %lld.%09ld\n\n", elapsed_t.tv_sec, elapsed_t.tv_nsec);
	if (elapsed_t.tv_sec < best_t[method].tv_sec || (elapsed_t.tv_sec == best_t[method].tv_sec && elapsed_t.tv_nsec < best_t[method].tv_nsec)) {
		best_t[method]=elapsed_t;
		}
	}

printf ("Raw counts from initial method, matched by all others:\n");
for (i=0; i<9; i++) {
	printf ("%d %d\n", i, known_good[i]);
	}
printf ("\nBest times for each method:\n");
for (i=0; i < NUMBER_OF_ALGOS; i++) {
	printf ("%20s: %lld.%09ld\n\n", method_names[i], best_t[i].tv_sec, best_t[i].tv_nsec);
	}

return (0);
}
