/*
 * Copyright 2023, Exotic Silicon, all rights reserved.
 *
 * This code is part of a programming tutorial.  For more information please visit:
 *
 * https://research.exoticsilicon.com/articles/utf8_parsing_in_c
 * gemini://gemini.exoticsilicon.com/articles/utf8_parsing_in_c
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. This software is licensed exclusively under this specific license text.  The license text may not be changed, and the software
 *    including modified versions may not be re-licensed under any other license text.
 * 2. Redistributions of source code must retain the above copyright notice, this list of conditions, and the following disclaimer.
 * 3. Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the following
 *    disclaimer in the documentation and/or other materials provided with the distribution.
 * 4. All advertising materials mentioning features or use of this software must display the following acknowledgement: This product
 *    includes software developed by Exotic Silicon.
 * 5. The name of Exotic Silicon must not be used to endorse or promote products derived from this software without specific prior
 *    written permission.
 * 6. Redistributions of modified versions of the source code must be clearly identified as having been modified from the original.
 * 7. Redistributions in binary form that have been created from modified versions of the source code must clearly state in the
 *    documentation and/or other materials provided with the distribution that the source code has been modified from the original.
 *
 * THIS SOFTWARE IS PROVIDED 'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EXOTIC SILICON BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Version 1.0
 */

#include <stdio.h>
#include <stdlib.h>

#define DEMO_SIZE 100000

#ifdef DEBUG
#define PRINTF printf
#else
#define PRINTF(...) {}
#endif

unsigned char * ERROR_STRINGS[] = {
    "OK", "Invalid byte", "Over-long encoding", "Surrogate half", "Codepoint above 0x10FFFF", "Unexpected non-continuation byte",
    "Unexpected continuation byte", "Input ends within multibyte token", "Memory allocation error", "Invalid parameters supplied"
    };

#define E_INVALID_BYTE			1
#define E_OVERLONG_ENCODING		2
#define E_SURROGATE			4
#define E_PAST_10FFFF			8
#define E_UNEXPECTED_SINGLE_OR_STARTER	16
#define E_UNEXPECTED_CONTINUATION_BYTE	32
#define E_END_WITHIN_TOKEN		64
#define E_MEMORY_ALLOCATION		128
#define E_INVALID_PARAMETERS		256

/*
 * Scan a string of length len bytes which is supplied in buffer and may contain embedded 0x00 bytes.
 * Return value indicates the first error found, or is zero if the string is strictly valid UTF-8.
 *
 *  1 Bit 0 - String contains an invalid byte such as 0xFE or 0xFF.
 *  2 Bit 1 - String contains an over-long encoding.
 *  4 Bit 2 - String contains an invalid codepoint, (surrogate half).
 *  8 Bit 3 - String contains an invalid codepoint, (> last char 0x10FFFF).
 * 16 Bit 4 - String contains non-continuation byte when continuation byte was expected.
 * 32 Bit 5 - String contains continuation byte when non-continuation byte was expected.
 * 64 Bit 6 - String ends in the middle of a multibyte character.
 */

int utf8_validate(unsigned char * buffer, unsigned int len, unsigned int * error_pos)
{
unsigned int i;
unsigned int codepoint_build;
unsigned char bitops;
unsigned char state; /* 0 if we are not already in a multibyte character, otherwise the number of bytes left */
unsigned char leading_ones;
unsigned char flag_byte_done;
unsigned char flag_overlong;
unsigned char multibyte_size;

#define CUR_BYTE *(buffer+i)
#define STORE_ERRPOS if (error_pos != NULL) { *error_pos = i; }

state=0;

for (i=0; i < len; i++) {
	flag_byte_done=0;
	bitops=CUR_BYTE;
	for (leading_ones=0; (bitops & 0x80)==0x80; leading_ones++, bitops=bitops << 1) ;
	if (state > 0 && leading_ones != 1) {
		STORE_ERRPOS
		PRINTF ("Non-continuation byte within multibyte sequence at string pos %u\n", i);
		return (16);
		}
	if (state > 0) {
		PRINTF ("Continuation byte at string pos %u\n", i);
		/*
		 * We can either write the lower six bits of the continuation byte directly to their correct location in codepoint_build
		 * or alternatively always write them to the lowest six bits of codepoint_build and shift whatever we already have left by
		 * six bits each time we process a continuation byte.  The second method is actually simpler and avoids multiplying by a
		 * value which is not a power of two.
		 */
		codepoint_build = codepoint_build << 6;
		codepoint_build |= CUR_BYTE & 0x3F;
		state--;
		flag_byte_done = 1;
		/*
		 * Finished decoding a multibyte?  Then check that the codepoint is valid.
		 */
		if (state==0) {
			/*
			 * Overly long encoding?
			 */
			flag_overlong=0;
			if (multibyte_size == 2 && codepoint_build < 0x80) {
				flag_overlong = 1;
				}
			if (multibyte_size == 3 && codepoint_build < 0x800) {
				flag_overlong = 1;
				}
			if (multibyte_size == 4 && codepoint_build < 0x10000) {
				flag_overlong = 1;
				}
			PRINTF ("%sultibyte, (%u bytes), codepoint 0x%x, (decimal %u) ending at string pos %u\n", (flag_overlong==1 ? "Overlong m" : "M"),
			    multibyte_size, codepoint_build, codepoint_build, i);
			if (flag_overlong == 1) {
				STORE_ERRPOS
				return (2);
				}
			if (codepoint_build > 0x10FFFF) {
				PRINTF ("Codepoints > 0x10FFFF are invalid.\n");
				STORE_ERRPOS
				return (8);
				}
			if (codepoint_build >= 0xD800 && codepoint_build <= 0xDFFF) {
				PRINTF ("Codepoint encodes a UTF-16 surrogate, and is therefore invalid in UTF-8.\n");
				STORE_ERRPOS
				return (4);
				}
			}
		}
	if (flag_byte_done == 0) {
		if (leading_ones == 1) {
			PRINTF ("Continuation byte without multibyte start byte\n");
			STORE_ERRPOS
			return (32);
			}
		if (leading_ones > 4) {
			PRINTF ("Invalid byte 0x%02x in UTF-8 stream at position %u\n", CUR_BYTE, i);
			STORE_ERRPOS
			return (1);
			}
		if (leading_ones > 1) {
			flag_byte_done = 1;
			state = multibyte_size = leading_ones;
			/*
			 * Decrement state by one, as the number of continuation bytes is one less than the total.
			 */
			state--;
			/*
			 * The leading ones are always followed by a zero.
			 * Therefore, the number of lower bits that contain codepoint data is (7 - leading_ones).
			 * We can create the necessary bitmask by right shifting 0xFF by (leading_ones) places.
			 */
			PRINTF ("Multibyte start at string pos %u\n", i);
			codepoint_build = (CUR_BYTE & (0xFF >> leading_ones));
			}
		}
	if (flag_byte_done == 0) {
		/*
		 * Must be a regular ASCII character
		 */
		PRINTF ("Regular ASCII character at string pos %u\n", i);
		flag_byte_done=1;
		}
	if (flag_byte_done == 0) {
		PRINTF ("Logic error in utf-8 parser!\n");
		exit(1);
		}
	}
if (state != 0) {
	PRINTF ("End of string within multibyte character - expecting %u more bytes\n", state);
	STORE_ERRPOS
	return (64);
	}
PRINTF ("Valid UTF-8 string of %u bytes\n", len);
return (0);
#undef CUR_BYTE
#undef STORE_ERRPOS
}

struct st_utf8_params {
	unsigned char * buffer_encoded;		/* Encoded data buffer */
	uint32_t * buffer_decoded;		/* Decoded data buffer */
	unsigned int buffer_encoded_len;	/* Encoded data buffer length */
	unsigned int buffer_decoded_len;	/* Decoded data buffer length, also used by decoder as a hint to size the output buffer if non-zero */
	unsigned int max_token_len;		/* Maximum token length to accept */
	unsigned int error_pos;			/* Position of fatal error */
	unsigned int ignore;			/* Bit flags for errors to ignore */
	unsigned int error_flags;		/* Bit flags for errors encountered and ignored */
	unsigned int error_len;			/* Number of bytes involved in an error, (used by pretty error printer) */
	};

/*
 * New error codes:
 * 128 Bit 7 - Memory allocation error.  (Cannot be ignored and will free buffer_decoded before returning)
 * 256 Bit 8 - Invalid parameters.  (Cannot be ignored and will free buffer_decoded before returning)
 */
int utf8_decode(struct st_utf8_params * params)
{
unsigned int i;
unsigned int codepoint_build;
unsigned int outbuf_size;
unsigned char bitops;
unsigned char state;
unsigned char leading_ones;
unsigned char flag_byte_done;
unsigned char flag_overlong;
unsigned char multibyte_size;
unsigned int * new_outbuf;

#define CUR_BYTE *(params->buffer_encoded + i)
#define STORE_ERRPOS { params->error_pos = i; }
#define MAX_OUTBUFFER_SIZE (256 * 1024 * 1024) /* 256 Mb */
#define REALLOC_INCREASE (64 * 1024) /* 64 Kb */

params->error_flags=0;
params->error_pos=0;
params->error_len=0;

if (params->max_token_len > 6 || params->max_token_len < 2 || params->buffer_encoded_len==0) {
	params->buffer_decoded_len=0;
	return (256);
	}

/*
 * We don't know in advance how many codepoints the input will decode to.
 * If params->buffer_decoded_len is non-zero, use that value as a hint to the expected number of codepoints and size the output buffer accordingly.
 * Otherwise size the output buffer to hold as many 32-bit codepoints as the input buffer has bytes, as the maximum possible number of output codepoints
 * occurs when the input is entirely ASCII with no multibyte tokens.
 */

#define INITIAL_SIZE (params->buffer_decoded_len == 0 ? params->buffer_encoded_len : params->buffer_decoded_len)
if (INITIAL_SIZE > MAX_OUTBUFFER_SIZE / sizeof(uint32_t)) {
	params->buffer_decoded_len=0;
	PRINTF ("Initial memory allocation exceeds maximum configured size of %u bytes\n", MAX_OUTBUFFER_SIZE);
	return (128);
	}
outbuf_size = sizeof(uint32_t) * INITIAL_SIZE;
#undef INITIAL_SIZE

PRINTF ("Allocating %u bytes for output buffer\n", outbuf_size);

params->buffer_decoded=malloc(outbuf_size);
if (params->buffer_decoded == NULL) {
	PRINTF ("Initial memory allocation error\n");
	params->buffer_decoded_len=0;
	return (128);
	}

state=0;
params->buffer_decoded_len=0;

for (i=0; i < params->buffer_encoded_len; i++) {
	if (params->buffer_decoded_len * sizeof(uint32_t)==outbuf_size) {
		if (outbuf_size + REALLOC_INCREASE > MAX_OUTBUFFER_SIZE) {
			free (params->buffer_decoded);
			params->buffer_decoded_len=0;
			params->error_len=1;
			STORE_ERRPOS
			return (128);
			}
		new_outbuf=realloc(params->buffer_decoded, outbuf_size + REALLOC_INCREASE);
		if (new_outbuf == NULL) {
			free (params->buffer_decoded);
			params->buffer_decoded_len=0;
			params->error_len=1;
			STORE_ERRPOS
			return (128);
			}
		PRINTF ("Reallocated output buffer from %u to %u\n", outbuf_size, outbuf_size + REALLOC_INCREASE);
		params->buffer_decoded=new_outbuf;
		outbuf_size += REALLOC_INCREASE;
		}
	flag_byte_done=0;
	bitops=CUR_BYTE;
	for (leading_ones=0; (bitops & 0x80)==0x80; leading_ones++, bitops=bitops << 1) ;
	if (state > 0 && leading_ones != 1) {
		PRINTF ("Non-continuation byte 0x%02x where continuation byte expected within multibyte sequence at string pos %u\n", CUR_BYTE, i);
		if ((params->ignore & 0x10) == 0) {
			params->error_len=1;
			STORE_ERRPOS
			return (16);
			}
		params->error_flags |= 0x10;
		/*
		 * If error 0x10 is being ignored then we discard the multi-byte sequence that was being constructed.
		 * Additionally, we maintain flag_byte_done=0 so that the code below will process it in some way.
		 * This could be as a regular ASCII character or as a new sequence introducer.
		 */
		state=0;
		}
	if (state > 0) {
		PRINTF ("Continuation byte 0x%02x at string pos %u\n", CUR_BYTE, i);
		codepoint_build = codepoint_build << 6;
		codepoint_build |= CUR_BYTE & 0x3F;
		state--;
		flag_byte_done = 1;
		/*
		 * Finished decoding a multibyte?  Then check that the codepoint is valid.
		 */
		if (state==0) {
			/*
			 * Overly long encoding?
			 */
			flag_overlong=0;
			if (codepoint_build < (multibyte_size == 2 ? 0x80 : (1 << (8 - multibyte_size + 6 * (multibyte_size - 2))))) {
				flag_overlong = 1;
				}
			PRINTF ("%sultibyte, (%u bytes), codepoint 0x%x, (decimal %u) ending at string pos %u\n", (flag_overlong==1 ? "Overlong m" : "M"),
			    multibyte_size, codepoint_build, codepoint_build, i);
			if (flag_overlong == 1) {
				if ((params->ignore & 0x02) == 0) {
					params->error_len=multibyte_size;
					STORE_ERRPOS
					return (2);
					}
				params->error_flags |= 0x02;
				}
			if (codepoint_build > 0x10FFFF) {
				PRINTF ("Codepoints > 0x10FFFF are invalid.\n");
				if ((params->ignore & 0x08) == 0) {
					params->error_len=multibyte_size;
					STORE_ERRPOS
					return (8);
					}
				params->error_flags |= 0x08;
				}
			if (codepoint_build >= 0xD800 && codepoint_build <= 0xDFFF) {
				PRINTF ("Codepoint encodes a UTF-16 surrogate, and is therefore invalid in UTF-8.\n");
				if ((params->ignore & 0x04) == 0) {
					params->error_len=multibyte_size;
					STORE_ERRPOS
					return (4);
					}
				params->error_flags |= 0x04;
				}
			/*
			 * Write the reconstructed codepoint to the output buffer.
			 * If errors 2, 4, or 8 are being ignored then it may have been constructed from invalid input, but we still write
			 * what we have to the output buffer and do _not_ discard it.
			 * If the desired behaviour is to discard invalid input, the calling function should not set the flags to ignore the
			 * corresponding error but instead allow this routine to exit early, then call it again with the start position after
			 * the non-conformant input byte.
			 */
			*(params->buffer_decoded+params->buffer_decoded_len++)=codepoint_build;
			}
		}
	if (flag_byte_done == 0) {
		if (leading_ones == 1) {
			PRINTF ("Continuation byte without multibyte start byte\n");
			if ((params->ignore & 0x20) == 0) {
				params->error_len=1;
				STORE_ERRPOS
				return (32);
				}
			params->error_flags |= 0x20;
			/*
			 * If error 0x20 is being ignored, then unexpected continuation bytes are silently dropped and the decoding continues.
			 */
			flag_byte_done=1;
			}
		if (leading_ones > params->max_token_len) {
			PRINTF ("Invalid byte 0x%02x in UTF-8 stream at position %u\n", CUR_BYTE, i);
			if ((params->ignore & 0x01) == 0) {
				params->error_len=1;
				STORE_ERRPOS
				return (1);
				}
			params->error_flags |= 0x01;
			/*
			 * If error 0x01 is being ignored, then a byte that would introduce a sequence longer than max_token_len is silently dropped
			 * and the decoding continues.  Any following continuation bytes that were part of such a longer sequence will be processed
			 * independently and the behaviour of the routine will depend on whether error 0x20 is being ignored or not, (see code above).
			 */
			flag_byte_done=1;
			}
		}
	if (flag_byte_done == 0) {
		if (leading_ones > 1) {
			flag_byte_done = 1;
			state = multibyte_size = leading_ones;
			/*
			 * Decrement state by one, as the number of continuation bytes is one less than the total.
			 */
			state--;
			/*
			 * The leading ones are always followed by a zero.
			 * Therefore, the number of lower bits that contain codepoint data is (7 - leading_ones).
			 * We can create the necessary bitmask by right shifting 0xFF by (leading_ones) places.
			 */
			PRINTF ("Multibyte start at string pos %u\n", i);
			codepoint_build = (CUR_BYTE & (0xFF >> leading_ones));
			}
		}
	if (flag_byte_done == 0) {
		/*
		 * Must be a regular ASCII character
		 */
		PRINTF ("Regular ASCII character at string pos %u\n", i);
		flag_byte_done=1;
		*(params->buffer_decoded+params->buffer_decoded_len++)=CUR_BYTE;
		}
	if (flag_byte_done == 0) {
		PRINTF ("Logic error in utf-8 parser!\n");
		exit(1);
		}
	}
if (state != 0) {
	PRINTF ("End of string within multibyte character - expecting %u more bytes\n", state);
	if ((params->ignore & 0x40) == 0) {
		params->error_len=multibyte_size-state;
		i--;
		STORE_ERRPOS
		return (64);
		}
	params->error_flags |= 0x40;
	}
PRINTF ("%salid UTF-8 string of %u bytes, decoded to %u codepoints\n", (params->error_flags == 0 ? "V" : "Parsable but strictly inv"),
    params->buffer_encoded_len, params->buffer_decoded_len);

#define OVER_ALLOC (outbuf_size - params->buffer_decoded_len * sizeof(uint32_t))
#define DECODED_BYTES (params->buffer_decoded_len * sizeof(uint32_t))
PRINTF ("Total output size is %lu, allocated buffer is %u.  Over-allocation was %lu bytes.\n", DECODED_BYTES, outbuf_size, OVER_ALLOC);
if (OVER_ALLOC > 32 * 1024) {
	PRINTF ("Reducing output buffer to %lu... ", DECODED_BYTES);
	new_outbuf=realloc(params->buffer_decoded, DECODED_BYTES);
	if (new_outbuf == NULL) {
		/*
		 * The re-allocation is always going to be a _reduction_, and never an increase.
		 * It's possible that on a memory-constrained system a large buffer could not be re-allocated and the call to realloc()
		 * might fail.  In this case we can still return normally because the original oversized buffer still exists, but we
		 * set bit 7 in the error flags to indicate that this non-critical error occurred.
		 */
		PRINTF ("failed!?\n");
		params->error_flags |= 128;
		return (0);
		}
	PRINTF ("OK!\n");
	params->buffer_decoded=new_outbuf;
	}
return (0);
#undef OVER_ALLOC
#undef CUR_BYTE
#undef STORE_ERRPOS
#undef MAX_OUTBUFFER_SIZE
#undef REALLOC_INCREASE
}


void utf8_dump_hex32(uint32_t * buffer, unsigned int len)
{
unsigned int i;

for (i=0; i < len; i++) {
	printf ("%06u 0x%08x\n", i, buffer[i]);
	}
return ;
}

int utf8_dump_error_bits(unsigned int errors)
{
unsigned int i;
unsigned int total;

total=0;
for (i=0x100; i > 0; i=i/2) {
	if ((errors & i)) {
		total++;
		}
	}
if (total == 0) {
	printf ("No errors were ignored.\n");
	return (0);
	}
printf ("The following error flags were set:\n");
printf ("Invalid byte                      %s\n", ((errors & 0x01) ? "YES" : "NO"));
printf ("Over-long encoding                %s\n", ((errors & 0x02) ? "YES" : "NO"));
printf ("Surrogate half                    %s\n", ((errors & 0x04) ? "YES" : "NO"));
printf ("Codepoint above 0x10FFFF          %s\n", ((errors & 0x08) ? "YES" : "NO"));
printf ("Unexpected non-continuation byte  %s\n", ((errors & 0x10) ? "YES" : "NO"));
printf ("Unexpected continuation byte      %s\n", ((errors & 0x20) ? "YES" : "NO"));
printf ("Input ends within multibyte token %s\n", ((errors & 0x40) ? "YES" : "NO"));
printf ("Memory allocation error           %s\n", ((errors & 0x80) ? "YES" : "NO"));
printf ("Invalid parameters supplied       %s\n", ((errors & 0x100) ? "YES" : "NO"));
printf ("Total types of errors ignored: %u\n", total);
return (total);
}

unsigned char * utf8_dump_error(unsigned int error)
{
unsigned int i;
unsigned int e;

e=9;
for (i=0x100; i > 0; i=i/2, e--) {
	if ((error & i)) {
		break ;
		}
	}
return (ERROR_STRINGS[e]);
}

void utf8_dump_error_pretty(struct st_utf8_params * params)
{
unsigned int i;

if (params->error_len==0) {
	return ;
	}
if (params->error_len > 8 || (params->error_pos < 8 && params->error_len > params->error_pos+1)) {
	printf ("Pretty error printer called with invalid parameters!\n");
	return ;
	}

#define FIELD_WIDTH 5
#define START (params->error_pos > 8 ? params->error_pos-8 : 0)
#define END (params->error_pos+8 > params->buffer_encoded_len ? params->buffer_encoded_len : params->error_pos+8)

for (i = START ; i < END ; i++) {
	printf ("  %c ", (params->buffer_encoded[i] >= 32 && params->buffer_encoded[i] < 128 ? params->buffer_encoded[i] : '?'));
	if (i + 1 < END) {
		printf (" ");
		}
	}
printf ("\n");

for (i = START ; i < END ; i++) {
	printf ("0x%02x", params->buffer_encoded[i]);
	if (i + 1 < END) {
		printf (" ");
		}
	}
printf ("\n");

for (i=0; i < FIELD_WIDTH * (params->error_pos >= 8 ? 9-params->error_len : params->error_pos+1-params->error_len); i++) {
	printf (" ");
	}

for (i=0; i <= (params->error_len-1); i++) {
	printf ("^^^^");
	if (i < (params->error_len-1)) {
		printf (" ");
		}
	}
printf ("\n");

#undef END
#undef START
#undef FIELD_WIDTH
}

/*
 * Allocate memory for a buffer of size codepoints, then fill it with mostly ASCII values and optionally a few higher codepoints.
 *
 * Flag bits:
 * 0 - 10FFFF / 0x7FFFFFFF maximum codepoint value, (if bit 2 is not also set).
 * 1 - Don't generate codepoints that represent UTF-16 surrogate halves / Do generate codepoints that represent UTF-16 surrogate halves.
 * 2 - Generate only ASCII values.
 */
int utf8_random_buffer(uint32_t ** buffer, unsigned int size, unsigned int flags)
{
#define FLAG_FULL_RANGE (flags & 0x01)
#define FLAG_INCLUDE_SURROGATES (flags & 0x02)
#define FLAG_ASCII_ONLY (flags & 0x04)
unsigned int * scratch_buffer;
unsigned int i;
unsigned int spos;
unsigned int flag_ok;

#define SCRATCH_ELEMENTS 65536
#define SCRATCH_SIZE (SCRATCH_ELEMENTS * sizeof(uint32_t))

scratch_buffer=malloc(SCRATCH_SIZE);
arc4random_buf(scratch_buffer, SCRATCH_SIZE);
spos=0;

printf ("Preparing %lu Kb of random%s data, (%u codepoints).\n", sizeof(uint32_t) * size / 1024, FLAG_ASCII_ONLY ? " ASCII-only" : "", size);
if (!FLAG_ASCII_ONLY) {
	printf ("Codepoints up to %s,%s including surrogates.\n", (FLAG_FULL_RANGE ? "0x7FFFFFFF" : "0x10FFFF"), (FLAG_INCLUDE_SURROGATES) ? "" : " not");
	}

*buffer=malloc(sizeof(uint32_t) * size);
i=0;
while (i < size) {
	if (FLAG_ASCII_ONLY ? 1 : arc4random_uniform(32) > 0) {
		(*buffer)[i]=scratch_buffer[spos++] & 0x7F;
		if (spos==SCRATCH_ELEMENTS) {
			arc4random_buf(scratch_buffer, SCRATCH_SIZE);
			spos=0;
			}
		if ((*buffer)[i] >= 32 && (*buffer)[i] < 127) {
			i++;
			}
		} else {
		flag_ok=0;
		while(flag_ok==0) {
			(*buffer)[i]=scratch_buffer[spos++] & 0x7FFFFFFF;
			if (spos==SCRATCH_ELEMENTS) {
				arc4random_buf(scratch_buffer, SCRATCH_SIZE);
				spos=0;
				}
			if (FLAG_FULL_RANGE || (*buffer)[i] <=0x10FFFF) {
				if ((*buffer)[i] >= 128 && (FLAG_INCLUDE_SURROGATES || ((*buffer)[i] < 0xd800 || (*buffer)[i] >0xdfff))) {
					i++;
					flag_ok=1;
					}
				}
			}
		}
	}
return (0);
#undef FLAG_FULL_RANGE
#undef FLAG_INCLUDE_SURROGATES
#undef FLAG_ASCII_ONLY
}

int utf8_encode(struct st_utf8_params * params)
{
#define MAX_OUTBUFFER_SIZE (256 * 1024 * 1024) /* 256 Mb */
#define	HIGH_BITS(i) ((0xFF << (8-i)) & 0xFF)
#define MAX_MAX_TOKEN_LEN 6
unsigned int cur_alloc;
unsigned int i;
unsigned int j;
unsigned int mb;
unsigned int new_alloc;
unsigned int max_codepoint[MAX_MAX_TOKEN_LEN + 1];
unsigned char * new_outbuf;

params->error_flags=0;
params->error_len=0; /* Fairly meaningless for the encoder as errors are always zero codepoints or one codepoint long. */
params->error_pos=0;
params->buffer_encoded_len=0;

/*
 * Check that the supplied parameters are sensible.
 */

if (params->buffer_decoded_len == 0 || params->max_token_len < 2 || params->max_token_len > MAX_MAX_TOKEN_LEN) {
	return (256);
	}

/*
 * Allocate sufficient memory for the maximum possible requirement.
 */

if (params->buffer_decoded_len > MAX_OUTBUFFER_SIZE / params->max_token_len) {
	return (128);
	}
cur_alloc=params->buffer_decoded_len * params->max_token_len;

params->buffer_encoded=malloc(cur_alloc);
if (params->buffer_encoded==NULL) {
	return (128);
	}

/*
 * Build a lookup table of multibyte character sizes as well as the correct bit shift and bit mask for the introducer.
 *
 * The max_codepoint array actually contains the _first invalid codepoint_ for the specified multibyte token length rather than the maximum valid one.
 * Note that uint32_t is sufficient to store the maximum value for a six-byte token, which is the largest commonly found even in sequences which are
 * not strictly UTF-8 compliant.  In the unlikely case of needing to process seven and eight byte tokens, then max_codepoint should be made uint64_t.
 * This would also require the constant 1 to be cast to a uint64_t: max_codepoint[i] = ((uint64_t)1 << (7-i + 6*(i-1)));
 */
for (i=2; i <= params->max_token_len; i++) {
	max_codepoint[i] = (1 << (7-i + 6*(i-1)));
	PRINTF ("%d byte sequence < %u, with overlay 0x%02x\n", i, max_codepoint[i], HIGH_BITS(i));
	}
max_codepoint[0]=0;
max_codepoint[1]=128;

#define CUR_CHAR params->buffer_decoded[i]

for (i=0; i < params->buffer_decoded_len; i++) {
	if (CUR_CHAR > 0x10FFFF) {
		if ((params->ignore & 0x08) == 0) {
			params->error_len=1;
			params->error_pos=i;
			return (8);
			}
		params->error_flags |= 0x08;
		}
	if (CUR_CHAR >= 0xD800 && CUR_CHAR <= 0xDFFF) {
		if ((params->ignore & 0x04) == 0) {
			params->error_len=1;
			params->error_pos=i;
			return (4);
			}
		params->error_flags |= 0x04;
		}
	for (j=1; j <= params->max_token_len; j++) {
		if (CUR_CHAR < max_codepoint[j]) {
			break ;
			}
		}
	if (j==params->max_token_len+1) {
		PRINTF ("Input codepoint 0x%x, (%u), exceeds maximum of 0x%x, (%u), encodable with %d bytes\n", CUR_CHAR, CUR_CHAR,
		    max_codepoint[params->max_token_len]-1, max_codepoint[params->max_token_len]-1, params->max_token_len);
		params->error_pos=i;
		params->error_len=1;
		return (1);
		}
	PRINTF ("Input codepoint of 0x%x, (%u), encodable with %u bytes: ", CUR_CHAR, CUR_CHAR, j);
	if (j==1) {
		params->buffer_encoded[params->buffer_encoded_len++]=CUR_CHAR;
		PRINTF ("%d\n", params->buffer_encoded_len);
		}
	if (j > 1) {
		params->buffer_encoded[params->buffer_encoded_len++]=(CUR_CHAR >> (6 * (j - 1))) | HIGH_BITS(j);
		PRINTF ("Bytes: %02x", params->buffer_encoded[params->buffer_encoded_len-1]);
		for (mb = j-1; mb > 0; mb--) {
			params->buffer_encoded[params->buffer_encoded_len++]=(((CUR_CHAR >> (6 * (mb - 1))) & 0x3f) | 0x80) ;
			PRINTF (" %02x", params->buffer_encoded[params->buffer_encoded_len-1]);
			}
		PRINTF ("\n");
		}
	}
#undef CUR_CHAR
#define OVER_ALLOC (cur_alloc - params->buffer_encoded_len)
PRINTF ("Total output size is %u, allocated buffer is %u.  Over-allocation was %u bytes.\n", params->buffer_encoded_len, cur_alloc, OVER_ALLOC);
if (OVER_ALLOC > 32 * 1024) {
	PRINTF ("Reducing output buffer to %u... ", params->buffer_encoded_len);
	new_outbuf=realloc(params->buffer_encoded, params->buffer_encoded_len);
	if (new_outbuf == NULL) {
		/*
		 * The re-allocation is always going to be a _reduction_, and never an increase.
		 * It's possible that on a memory-constrained system a large buffer could not be re-allocated and the call to realloc()
		 * might fail.  In this case we can still return normally because the original oversized buffer still exists, but we
		 * set bit 7 in the error flags to indicate that this non-critical error occurred.
		 */
		PRINTF ("failed!?\n");
		params->error_flags |= 128;
		return (0);
		}
	PRINTF ("OK!\n");
	params->buffer_encoded=new_outbuf;
	}
#undef OVER_ALLOC
#undef HIGH_BITS
#undef MAX_OUTBUFFER_SIZE
return (0);
}

int utf8_compare(unsigned int * buffer_a, unsigned int * buffer_b, unsigned int size)
{
unsigned int i;
unsigned int errors;

errors=0;
for (i=0; i<size; i++) {
	if (buffer_a[i] != buffer_b[i]) {
		printf ("Mismatch at offset %u, (%u / %u)\n", i, buffer_a[i], buffer_b[i]);
		errors++;
		if (errors > 20) {
			printf ("More than 20 mismatched bytes, stopping here.\n");
			return (0xFF);
			}
		}
	}
return (errors);
}

int utf8_demo()
{
unsigned int res;
uint32_t * original_data;
struct st_utf8_params params;

utf8_random_buffer(&params.buffer_decoded, DEMO_SIZE, 0);
printf ("\nStarting encoding\n");

params.buffer_decoded_len=DEMO_SIZE;	/* Size of input */
params.buffer_encoded_len=0;		/* Optional hint for the output buffer size */
params.max_token_len=6;
params.ignore=0;

res=utf8_encode(&params);

printf ("Encoder returned: %s\n", utf8_dump_error(res));
printf ("Encoded length: %u\n", params.buffer_encoded_len);
utf8_dump_error_bits(params.error_flags);

if (res != 0) {
	printf ("Encoder exited on error.\n");
	return (0xfe);
	}

original_data=params.buffer_decoded;
params.buffer_decoded=NULL;
params.buffer_decoded_len=0;

#ifdef OVERWRITE_A_RANDOM_BYTE
params.buffer_encoded[arc4random_uniform(params.buffer_encoded_len)] = arc4random_uniform(256);
#endif

printf ("\nStarting validation\n");
res=utf8_validate(params.buffer_encoded, params.buffer_encoded_len, NULL);
printf ("Validator returned: %s\n", utf8_dump_error(res));

printf ("\nStarting decoding\n");
res=utf8_decode(&params);

printf ("Decoder returned: %s\n", utf8_dump_error(res));
printf ("Decoded length: %u\n", params.buffer_decoded_len);
utf8_dump_error_pretty(&params);
utf8_dump_error_bits(params.error_flags);

if (res != 0) {
	printf ("Decoder exited on error.\n");
	return (0xfd);
	}

printf ("\nComparing original data with encoded and decoded copy...\n");

res=utf8_compare(original_data, params.buffer_decoded, DEMO_SIZE);
printf ("%s\n", (res==0 ? "Success!" : "Failed."));

return (res);
}

int main()
{
utf8_demo();
return (0);
}
